function testFTSensor(port, baudrate, timeInMilliSeconds)

    %PARSE INPUTS
    if(~exist('port','var') || isempty(port))
        port = '/dev/ttyS0';
    end
    
    if(~exist('baudrate','var') || isempty(baudrate))
        baudrate = 57600; %use, e.g., cutecom to set the baudrate on the sensor side (send the string: v057600) 
    end
    
    if(~exist('timeInMilliSeconds','var') || isempty(timeInMilliSeconds))
        timeInMilliSeconds = 20; %cyclic data event invoke interval
    end

    %ATTENTION: If you receive the message Could not find port ...
    %run sudo rm /var/lock/LCK..ttyS0
    
    if 1==0
        %DIRECT JAVA CLASS ACCESS
        import de.IAS.FTCL5040Sensor;
        schunkSensor = FTCL5040Sensor(port, baudrate);
        pause(2)
        schunkSensor.open();
        pause(2)

        schunkSensor.setSensorToZero();

        result = schunkSensor.readForcesAndTorquesBlocking()    %single blocking read


        schunkSensor.close();
    else
        
         
        schunkSensor = MatlabFTCL5040Sensor(port, baudrate);
        schunkSensor.open();
        %schunkSensor.setSensorToZero();
        fprintf('\n\nStarting single blocking read ... \n');
        result = schunkSensor.readForcesAndTorquesBlocking()    %single blocking read
        
       
        fprintf('\n\nStart Cyclic read ... \n\n');
        schunkSensor.startCyclicTransmission(timeInMilliSeconds);

        fprintf('\n\nPrinting Force Values\n');
        schunkSensor.addCallback(@printForceValues);
        pause(2);
        schunkSensor.callbacks = {};
         
        fprintf('\n\nPlotting Froce Values\n');
        schunkSensor.addCallback(@plotForceValues);
        pause(10);
        schunkSensor.callbacks = {};
         
        fprintf('\n\nChecking frequency\n');
        schunkSensor.addCallback(@frequencyCheck);
        pause(2)
        schunkSensor.callbacks = {};
     

        schunkSensor.close();
        schunkSensor.delete();
    end
   
    

end


%HELPER
 function frequencyCheck( values )
    persistent t1;
    if(isempty(t1))
        t1 = tic;
    else
        toc(t1);
        t1 = tic;
    end
end

function printForceValues( values )
    fprintf('VALUES ');
    fprintf(' %.4f',values);
    fprintf('\n');
end

function plotForceValues( values )
    bar(values);
    ylim([-100 100]);
end
