//to compile run javac -cp /usr/share/java/RXTXcomm.jar Sensor.java
//               jar cf SchunkSensor.jar Sensor.class
//Elmar Rueckert, rueckert.elmar@gmail.com
//Sept, 2015

package de.IAS;

import gnu.io.*; // RXTX

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.TooManyListenersException;

/**
 * @author rueckert
 *
 */
public class FTCL5040Sensor extends Thread {

	//MEMBERS
	CommPortIdentifier serialPortId;
	Enumeration<?> enumComm;
	SerialPort serialPort;
	InputStream inputStream;
	OutputStream outputStream;
	Boolean serialPortOpen = false;
	float[] lastForceTorqueValues = new float[6];
	byte[] byteBuffer = new byte[500];
	boolean flgFoundStartPattern = false;
	int patternPtr = 0;
	
//	int factoryBaudrate = 9600;
	int runtimeBaudrate = 9600;
	int dataBits = SerialPort.DATABITS_8;
	int stopBits = SerialPort.STOPBITS_1;
	int parity = SerialPort.PARITY_NONE;
	int flowControl = SerialPort.FLOWCONTROL_NONE;
	boolean flgStreamingIsActivated = false;

	String portName = "";
	private final PropertyChangeSupport pcs = new PropertyChangeSupport(this);

	//NESTED CLASS
	class serialPortEventListener implements SerialPortEventListener {
		public void serialEvent(SerialPortEvent event) {
			switch (event.getEventType()) {
			case SerialPortEvent.DATA_AVAILABLE:
				serialPortDataReady();
				break;
			case SerialPortEvent.BI:
				break;
			case SerialPortEvent.CD:
				break;
			case SerialPortEvent.CTS:
				break;
			case SerialPortEvent.DSR:
				break;
			case SerialPortEvent.FE:
				break;
			case SerialPortEvent.OUTPUT_BUFFER_EMPTY:
				break;
			case SerialPortEvent.PE:
				break;
			case SerialPortEvent.RI:
				break;
			default:
			}
		}
	}

	//Constructor
	public FTCL5040Sensor(String portName){
		this.portName = portName;
	}
	public FTCL5040Sensor(String portName, int baudrate){
		this.portName = portName;
		if(baudrate != 1200 && baudrate != 2400 &&
				baudrate != 9600 && baudrate != 19200 &&
				baudrate != 38400 && baudrate != 57600 && baudrate != 115200 && 
				baudrate != 230400 && baudrate != 460800 && baudrate != 921600) {
			System.out.println("Invalid argument: use a valid baudrate, e.g., 57600");
			return;
		} else {
			this.runtimeBaudrate = baudrate;
		}
	}

	//PUBLIC FUNCTIONS
	public void open(){
		super.start();
//		try{
//			Thread.sleep(500);
//		}catch(Exception e){
//			e.printStackTrace();
//		}
	}

	public void close(){
		if ( serialPortOpen == true) {
			
			stopCyclicTransmission();
			
//			if(this.runtimeBaudrate != this.factoryBaudrate) {
//				setBaudrate(factoryBaudrate);
//			}
			
			serialPort.close();
			serialPortOpen = false;
			try {
				Thread.sleep(1000);
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		} else {
			System.out.println("Serialport already closed");
		}
	}

	@Override
	public void run(){
		openSerialPort(this.portName);
		
//		if(this.runtimeBaudrate != this.factoryBaudrate) {
//			setBaudrate(this.runtimeBaudrate);
//		}
	}

	public void setSensorToZero(){
		if(sendCommand('z', '{') == false) {
			System.out.println("Could not set the sensor to zero");
		}
	}

	public void resetSensor(){
		if(sendCommand('0', '1') == false) {
			System.out.println("Could not reset the sensor");
		}
		flgStreamingIsActivated = false;
	}

	public void startCyclicTransmission(){ startCyclicTransmission(10); } //100Hz streaming
	public void startCyclicTransmission(int timeInMilliSeconds){

		try{
			Thread.sleep(500);
		}catch(Exception e){
			e.printStackTrace();
		}

		setDataTransferTime(timeInMilliSeconds);
		
		try{
			Thread.sleep(500);
		}catch(Exception e){
			e.printStackTrace();
		}
		
		sendCommand("D2");
		
//		try{
//			Thread.sleep(200);
//		}catch(Exception e){
//			e.printStackTrace();
//		}
		//		try {
		//		//setDataTransferTime(timeInMilliSeconds);
		//
		//		sendCommand("D2");
		//		
		//			Thread.sleep(1);
		//		} catch (InterruptedException e) {
		//			// TODO Auto-generated catch block
		//			e.printStackTrace();
		//		}
//		System.out.println("Streaming is activated");
		//		try{
		//			Thread.sleep(500);
		//		}catch(Exception e){
		//			e.printStackTrace();
		//		}
		flgStreamingIsActivated = true;
	}

	public void stopCyclicTransmission(){
		if(flgStreamingIsActivated==true){
			sendCommand("4");
//			System.out.println("Streaming is deactivated");
			flgStreamingIsActivated = false;
			try {
				Thread.sleep(500);
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}

	public void addPropertyChangeListener(PropertyChangeListener listener) {
		this.pcs.addPropertyChangeListener(listener);
	}

	public void removePropertyChangeListener(PropertyChangeListener listener) {
		this.pcs.removePropertyChangeListener(listener);
	}

	public int[] readTestIntValuesBlocking() throws Exception {
		if (serialPortOpen != true)
			throw new Exception("Serial Port is not Open!");
		try {
			int[] forceValues = new int[6];
			byte[] data = new byte[150];
			int num = 0;

			//empty buffer
			while(inputStream.available() > 0) {
				num = inputStream.read(data, 0, data.length);
				Thread.sleep(10);//sleep
			}
			//initiate data transfer
			outputStream.write(("N").getBytes());

			int numTrials = 0;
			int numExpectedBytes = 15;

			while(inputStream.available() < numExpectedBytes && numTrials++ < 100) { 
				Thread.sleep(10);//sleep
			}
			while(inputStream.available() > 0) {
				num = inputStream.read(data, 0, data.length);
			}
			//O int1L int1H int2L int2H int3L int3H int4L int4H int5L int5H int6L int6H intStatusL intStatusH 
			if(num >= numExpectedBytes) {
				//found a pattern to parse
				forceValues[0] = int16FromByteArray(Arrays.copyOfRange(data, 1, 3));
				forceValues[1] = int16FromByteArray(Arrays.copyOfRange(data, 3, 5));
				forceValues[2] = int16FromByteArray(Arrays.copyOfRange(data, 5, 7));
				forceValues[3] = int16FromByteArray(Arrays.copyOfRange(data, 7, 9));
				forceValues[4] = int16FromByteArray(Arrays.copyOfRange(data, 9, 11));
				forceValues[5] = int16FromByteArray(Arrays.copyOfRange(data, 11, 13));

				return forceValues;
			}

		} catch (IOException e) {
			System.out.println("Could not send data to serialport");
		} catch (InterruptedException e) {
			System.out.println("Could not send data to serialport");
		}
		return null;
	}

	public float[] readTestFloatValuesBlocking() throws Exception {
		if (serialPortOpen != true)
			throw new Exception("Serial Port is not Open!");
		try {
			float[] forceValues = new float[6];
			byte[] data = new byte[150];
			int num = 0;

			//empty buffer
			while(inputStream.available() > 0) {
				num = inputStream.read(data, 0, data.length);
				Thread.sleep(10);//sleep
			}
			//initiate data transfer
			outputStream.write(("F").getBytes());

			int numTrials = 0;
			int numExpectedBytes = 27;

			while(inputStream.available() < numExpectedBytes && numTrials++ < 100) { 
				Thread.sleep(10);//sleep
			}
			while(inputStream.available() > 0) {
				num = inputStream.read(data, 0, data.length);
			}
			//O int1L int1H int2L int2H int3L int3H int4L int4H int5L int5H int6L int6H intStatusL intStatusH 
			if(num >= numExpectedBytes) {
				//found a pattern to parse
				forceValues[0] = floatFromByteArray(Arrays.copyOfRange(data, 1, 5)); 
				forceValues[1] = floatFromByteArray(Arrays.copyOfRange(data, 5, 9));
				forceValues[2] = floatFromByteArray(Arrays.copyOfRange(data, 9, 13));
				forceValues[3] = floatFromByteArray(Arrays.copyOfRange(data, 13, 17));
				forceValues[4] = floatFromByteArray(Arrays.copyOfRange(data, 17, 21));
				forceValues[5] = floatFromByteArray(Arrays.copyOfRange(data, 21, 25));

				return forceValues;
			}

		} catch (IOException e) {
			System.out.println("Could not send data to serialport");
		} catch (InterruptedException e) {
			System.out.println("Could not send data to serialport");
		}
		return null;
	}

	public int[] readIntForceValuesBlocking() throws Exception {
		if (serialPortOpen != true)
			throw new Exception("Serial Port is not Open!");
		try {
			int[] returnValues = new int[3];
			outputStream.write(("H").getBytes());
			byte[] data = new byte[150];
			int num = 0;
			int numTrials = 0;
			int numExpectedBytes = 9;

			while(inputStream.available() < numExpectedBytes && numTrials++ < 100) { //inputStream.available() > 0) {
				Thread.sleep(1);//sleep
			}
			while(inputStream.available() > 0) {
				num = inputStream.read(data, 0, data.length);
			}
			//I int1L int1H int2L int2H int3L int3H intStatusL intStatusH 
			if(num >= numExpectedBytes) {
				//found a pattern to parse
				returnValues[0] = int16FromByteArray(Arrays.copyOfRange(data, 1, 3));
				returnValues[1] = int16FromByteArray(Arrays.copyOfRange(data, 3, 5));
				returnValues[2] = int16FromByteArray(Arrays.copyOfRange(data, 5, 7));
				return returnValues;
			}

		} catch (IOException e) {
			System.out.println("Could not send data to serialport");
		} catch (InterruptedException e) {
			System.out.println("Could not send data to serialport");
		}
		return null;
	}

	public float[] readForcesAndTorquesBlocking() {
		if (serialPortOpen != true){
			System.out.println("Serial Port is not Open!"); 
			return null;
		}
		try {
			outputStream.write(("D").getBytes());
			byte[] data = new byte[150];
			int num = 0;
			int numTrials = 0;
			int numExpectedBytes = 27;

			while(inputStream.available() < numExpectedBytes && numTrials++ < 100) { //inputStream.available() > 0) {
				Thread.sleep(1);//sleep
			}
			while(inputStream.available() > 0) {
				num = inputStream.read(data, 0, data.length);
			}
			
//			System.out.println("Received blocking: " + num + " bytes");
			
			//I int1L int1H int2L int2H int3L int3H intStatusL intStatusH 
			if(num >= numExpectedBytes) {
				//found a pattern to parse
				lastForceTorqueValues[0] = floatFromByteArray(Arrays.copyOfRange(data, 1, 5)); 
				lastForceTorqueValues[1] = floatFromByteArray(Arrays.copyOfRange(data, 5, 9));
				lastForceTorqueValues[2] = floatFromByteArray(Arrays.copyOfRange(data, 9, 13));
				lastForceTorqueValues[3] = floatFromByteArray(Arrays.copyOfRange(data, 13, 17));
				lastForceTorqueValues[4] = floatFromByteArray(Arrays.copyOfRange(data, 17, 21));
				lastForceTorqueValues[5] = floatFromByteArray(Arrays.copyOfRange(data, 21, 25));
				return lastForceTorqueValues;
			}

		} catch (IOException e) {
			System.out.println("Could not send data to serialport");
		} catch (InterruptedException e) {
			System.out.println("Could not send data to serialport");
		}
		return null;
	}

	//	//PROTECTED FUNCTIONS
	//	protected void serialPortDataReady() {
	//
	//		if(flgStreamingIsActivated == false){
	//			return;
	//		}
	//
	//		try {
	//			byte[] data = new byte[150];
	//			int num = 0;
	//			while(inputStream.available() > 0) {
	//				num = inputStream.read(data, 0, data.length);
	//			}
	//			int numExpectedBytes = 27;
	//			if(num>=1){//numExpectedBytes){
	//				System.out.println("Data received " + num);
	//			}
	//			numExpectedBytes = 0;
	//			//			
	//			//			int[] flexValues = new int[5];
	//			//			int idx = 0;
	//			//			boolean found = false;
	//			//			for( idx = 1; idx <= (num-numExpectedBytes); idx ++){
	//			//
	//			//				if( (data[idx] == 114)
	//			//						&& (data[idx+3] == 44)
	//			//						&& (data[idx+6] == 44)
	//			//						&& (data[idx+9] == 44)
	//			//						&& (data[idx+12] == 44)
	//			//						&& (data[idx+15] == 44)
	//			//						&& (data[idx+16] == 120)
	//			//						){
	//			//					flexValues[0] = ((0xFF & data[idx+1]) << 8) | (0xFF & data[idx+2]);
	//			//					flexValues[1] = ((0xFF & data[idx+4]) << 8) | (0xFF & data[idx+5]);
	//			//					flexValues[2] = ((0xFF & data[idx+7]) << 8) | (0xFF & data[idx+8]);
	//			//					flexValues[3] = ((0xFF & data[idx+10]) << 8) | (0xFF & data[idx+11]);
	//			//					flexValues[4] = ((0xFF & data[idx+13]) << 8) | (0xFF & data[idx+14]);
	//			//
	//			//					found = true;
	//			//					break;
	//			//					//this.pcs.firePropertyChange("sensorFlexValues", 0, flexValues);
	//			//				}
	//			//
	//			//			}
	//			//
	//			//			if(found){
	//			//				this.pcs.firePropertyChange("sensorValues", 0, flexValues);
	//			//			}
	//
	//		} catch (IOException e) {
	//			System.out.println("Could not read data from serialport");
	//		}
	//	}


	//PROTECTED FUNCTIONS
	protected void serialPortDataReady() {

		
		if(flgStreamingIsActivated==false){ return; }

		try {

			byte nextByte;
			int numExpectedBytes = 27;
			
			while(inputStream.available() > 0) {
				nextByte = (byte)inputStream.read();
				
				if(flgFoundStartPattern==false && nextByte == 69){ //'E'
					flgFoundStartPattern = true;
					patternPtr=0;
//					System.out.println("Start Received");
				}
				
				if(flgFoundStartPattern==true){
					byteBuffer[patternPtr++] = nextByte;
					if(patternPtr>=numExpectedBytes-1){ break; }
				}
			}
			if(flgFoundStartPattern==false){
				return;
			}
			
//			System.out.println(patternPtr);
			
			if(flgFoundStartPattern == true && patternPtr >= numExpectedBytes-1){
				
//				System.out.println("p");
//				
//				lastForceTorqueValues[0] = floatFromByteArray(Arrays.copyOfRange(byteBuffer, 1, 5));
//				lastForceTorqueValues[1] = floatFromByteArray(Arrays.copyOfRange(byteBuffer, 5, 9));
//				lastForceTorqueValues[2] = floatFromByteArray(Arrays.copyOfRange(byteBuffer, 9, 13));
//				lastForceTorqueValues[3] = floatFromByteArray(Arrays.copyOfRange(byteBuffer, 13, 17));
//				lastForceTorqueValues[4] = floatFromByteArray(Arrays.copyOfRange(byteBuffer, 17, 21));
//				lastForceTorqueValues[5] = floatFromByteArray(Arrays.copyOfRange(byteBuffer, 21, 25));
				
				lastForceTorqueValues[0] = floatFromByteArray(byteBuffer[1],byteBuffer[2],byteBuffer[3],byteBuffer[4]); 
				lastForceTorqueValues[1] = floatFromByteArray(byteBuffer[5],byteBuffer[6],byteBuffer[7],byteBuffer[8]);
				lastForceTorqueValues[2] = floatFromByteArray(byteBuffer[9],byteBuffer[10],byteBuffer[11],byteBuffer[12]);
				lastForceTorqueValues[3] = floatFromByteArray(byteBuffer[13],byteBuffer[14],byteBuffer[15],byteBuffer[16]);
				lastForceTorqueValues[4] = floatFromByteArray(byteBuffer[17],byteBuffer[18],byteBuffer[19],byteBuffer[20]);
				lastForceTorqueValues[5] = floatFromByteArray(byteBuffer[21],byteBuffer[22],byteBuffer[23],byteBuffer[24]);
				
				
				
				patternPtr = 0;
				flgFoundStartPattern = false;
				
				this.pcs.firePropertyChange("ForceTorqueValues", 0, lastForceTorqueValues);
			}
			
		} catch (IOException e) {
			System.out.println("Could not read data from serialport");
		}
	}

	protected boolean openSerialPort(String portName){

		Boolean foundPort = false;
		if (serialPortOpen != false) {
			System.out.println("Port already open!");
			return false;
		}

		enumComm = CommPortIdentifier.getPortIdentifiers();
		while(enumComm.hasMoreElements()) {
			serialPortId = (CommPortIdentifier) enumComm.nextElement();

			//             if(serialPortId.getPortType() == CommPortIdentifier.PORT_SERIAL) {
			//               System.out.println("Found serial port: " + serialPortId.getName());
			//             }

			if (portName.contentEquals(serialPortId.getName())) {
				foundPort = true;
				break;
			}
		}

		if (foundPort != true) {
			System.out.println("Could not find port: " + portName);
			return false;
		}

		try {
			serialPort = (SerialPort) serialPortId.open("Öffnen und Senden", 500);
		} catch (PortInUseException e) {
			System.out.println("Port already in use.");
		}

		try {
			outputStream = serialPort.getOutputStream();
		} catch (IOException e) {
			System.out.println("Can not access output stream.");
		}

		try {
			inputStream = serialPort.getInputStream();
		} catch (IOException e) {
			System.out.println("Can not access input stream.");
		}
		try {
			serialPort.addEventListener(new serialPortEventListener());
		} catch (TooManyListenersException e) {
			System.out.println("TooManyListenersException");
		}
		serialPort.notifyOnDataAvailable(true);
		try {
			serialPort.setSerialPortParams(this.runtimeBaudrate, dataBits, stopBits, parity);
			serialPort.setFlowControlMode(flowControl);
		} catch(UnsupportedCommOperationException e) {
			System.out.println("Could not set port parameters");
		}

		try {
			Thread.sleep(1000);
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		serialPortOpen = true;
		return true;
	}

	protected int int16FromByteArray(byte[] bytes) {
		return bytes[1] << 8 | (bytes[0] & 0xFF);
	}
	protected float floatFromByteArray(byte[] bytes) {
		int tmp = bytes[3] << 24 | (bytes[2] & 0xFF) << 16 | (bytes[1] & 0xFF) << 8 | (bytes[0] & 0xFF);
		float retValue = Float.intBitsToFloat(tmp);
		return retValue;
	}
	protected float floatFromByteArray(byte b0, byte b1, byte b2, byte b3) {
		int tmp = b3 << 24 | (b2 & 0xFF) << 16 | (b1 & 0xFF) << 8 | (b0 & 0xFF);
		float retValue = Float.intBitsToFloat(tmp);
		return retValue;
	}

	protected boolean sendCommand(Character command, Character expectedResult){
		return sendCommand(command.toString(), expectedResult);
	}
	protected boolean sendCommand(String command, Character expectedResult){
		try {
			outputStream.write(command.getBytes());
			int numTrials = 0;
			int numExpectedBytes = 1;
			while(inputStream.available() < numExpectedBytes && numTrials++ < 100) { 
				Thread.sleep(1);//sleep
			}
			if(inputStream.available() > 0) {
				int result = (int)inputStream.read();
				if(result == Integer.valueOf(expectedResult)) {
					return true;
				} else {
					System.out.println("Sensor Error Code: " + result);
				}
			}
		}catch(Exception e) {
			System.out.println("Internal Error: " + e.getMessage());
		}
		return false;
	}

	protected void sendCommand(String command){
		try {
			outputStream.write(command.getBytes());
			
//			int numExpectedBytes = 3;
//			int numTrials = 0;
//			while(inputStream.available() < numExpectedBytes && numTrials++ < 100) { 
//				Thread.sleep(10);//sleep
//			}
//			while(inputStream.available() > 0) {
//				inputStream.read();
//			}
			
		}catch(Exception e) {
			System.out.println("Failed to send a command without response");
		}
	}
	protected void setDataTransferTime(int timeInMilliSeconds){

		if(timeInMilliSeconds < 1 || timeInMilliSeconds > 65767) {
			System.out.println("Invalid argument: timeInMilliSeconds must be in the range [1,65767]");
			return;
		}

		if(sendCommand('&', '\'') == false) {
			System.out.println("Could not set the data transfer rate");
			return;
		}
		String msAsString = String.format("%05d", timeInMilliSeconds);
		sendCommand(msAsString);
	}
	protected void setBaudrate(int baudrate){

		String msAsString = String.format("%06d", baudrate);
		boolean retErr = sendCommand("v" + msAsString, 'w');
		if(retErr==false){
			System.out.println("FATAL ERROR! RESET HARDWARE");
		}else {
			try {
				serialPort.setSerialPortParams(baudrate, dataBits, stopBits, parity);
			} catch(UnsupportedCommOperationException e) {
				System.out.println("Could not change port parameters");
			}
		}
		try {
			System.out.println("sensor is rebooting, wait for 5 seconds ... ");
			Thread.sleep(5000);
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	//	protected byte[] int16ToByteArray(int value) {
	//		return new byte[] { 
	//				(byte)value , (byte)(value >> 8)
	//				};
	//	}
	//	protected byte[] floatToByteArray(float inputValue) {
	//		int value = Integer.
	//		return new byte[] { 
	//				(byte)(value),
	//		        (byte)(value >> 8),
	//		        (byte)(value >> 16),
	//		        (byte)(value >> 24)};
	//	}

}
